// Debugger.cpp : implementation file
//

#include "stdafx.h"
#include "breem.h"
#include "Debugger.h"
#include "dead6502\dead6502.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define LINES	18

/////////////////////////////////////////////////////////////////////////////
// CDebugger dialog


CDebugger::CDebugger(CWnd* pParent /*=NULL*/)
	: CDialog(CDebugger::IDD, pParent)
{
	//{{AFX_DATA_INIT(CDebugger)
	m_stepcycles = _T("1");
	m_jumpaddr = _T("$c000");
	m_info = _T("");
	m_breakaddr = _T("$c000");
	//}}AFX_DATA_INIT
currentcpu = (u8)-1;
up.LoadBitmap(IDB_UP);
down.LoadBitmap(IDB_DOWN);
}


void CDebugger::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CDebugger)
	DDX_Control(pDX, IDC_DISASM, m_disasm);
	DDX_Control(pDX, IDC_UP, m_up);
	DDX_Control(pDX, IDC_DOWN, m_down);
	DDX_Control(pDX, IDC_CPUTABS, m_tabs);
	DDX_Text(pDX, IDC_STEPCYCLES, m_stepcycles);
	DDX_Text(pDX, IDC_JUMPADDR, m_jumpaddr);
	DDX_Text(pDX, IDC_INFO, m_info);
	DDX_Text(pDX, IDC_BREAKADDR, m_breakaddr);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CDebugger, CDialog)
	//{{AFX_MSG_MAP(CDebugger)
	ON_NOTIFY(TCN_SELCHANGE, IDC_CPUTABS, OnSelchangeCputabs)
	ON_BN_CLICKED(IDC_REFRESH, OnRefresh)
	ON_BN_CLICKED(IDC_UP, OnUp)
	ON_BN_CLICKED(IDC_DOWN, OnDown)
	ON_BN_CLICKED(IDC_STEP, OnStep)
	ON_BN_CLICKED(IDC_RESET, OnReset)
	ON_WM_CLOSE()
	ON_BN_CLICKED(IDC_BREAK, OnBreak)
	ON_BN_CLICKED(IDC_JUMP, OnJump)
	ON_BN_CLICKED(IDC_IRQ, OnIrq)
	ON_BN_CLICKED(IDC_NMI, OnNmi)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

void CDebugger::setcontexts(dead6502data *cpu1,dead6502data *cpu2,romset_t *roms)
{
maincpu = cpu1;
soundcpu = cpu2;
romset = roms;
cpu1frame = SCANLINECYCLES(romset->cpu1hz) * 272;
cpu2frame = SCANLINECYCLES(romset->cpu2hz) * 272;
cpu1screen = SCANLINECYCLES(romset->cpu1hz) * 240;
cpu2screen = SCANLINECYCLES(romset->cpu2hz) * 240;
cpu1cycles = 0;
cpu2cycles = 0;
}

char *getflags(u8 flags,char *s)
{
#define FLAG(v,l)	flags&v?l:'-'
sprintf(s,"%c%c%c%c%c%c%c",FLAG(0x80,'S'),FLAG(0x40,'V'),FLAG(0x10,'B'),FLAG(0x08,'D'),FLAG(0x04,'I'),FLAG(0x02,'Z'),FLAG(0x01,'C'));
return(s);
}

void CDebugger::refresh()
{
char str[128];
int i,j,sl;
u16 p;
int line;

if(currentcpu == 0) //main
	line = (cpu1cycles / SCANLINECYCLES(romset->cpu1hz)) + 1;
else
	line = (cpu2cycles / SCANLINECYCLES(romset->cpu2hz)) + 1;
m_info.Format("%04X\n%02X\n%02X\n%02X\n%02X\n%02X\n%s\n%ld\n\n%04X\n%04X\n%04X\n\n%d",cpu->pc,cpu->a,cpu->x,cpu->y,cpu->s,cpu->f,getflags(cpu->f,str),cpu->totalcycles,getvector(NMI),getvector(IRQ),getvector(RESET),line);
m_disasm.ResetContent();
p = pc;
line = -1;
for(i=0;i<LINES;i++)
	{
	if(p == cpu->pc)
		line = i;
	p = disassembleopcode(str,p);
	sl = strlen(str);
	for(j=0;j<sl;j++)
		{
		if(str[j] == '\n')
			str[j] = ' ';
		}
	m_disasm.AddString(str);
	}
if(line != -1)
	m_disasm.SetCurSel(line);
UpdateData(FALSE);
}

void CDebugger::changecpu()
{
if(currentcpu == 0) //main cpu
	cpu = maincpu;
else if(currentcpu == 1) //sound cpu
	cpu = soundcpu;
else //error
	{
	::MessageBox(0,"bug: currentcpu isnt 0 or 1 so im setting it to 0 now (the main cpu), email deadbody@parodius.com and tell me im a bad coder","error",MB_OK);
	currentcpu = 0;
	cpu = maincpu;
	}
set6502data(cpu);
pc = cpu->pc;
}

/////////////////////////////////////////////////////////////////////////////
// CDebugger message handlers

BOOL CDebugger::OnInitDialog() 
{
	CDialog::OnInitDialog();
m_up.SetBitmap(up);
m_down.SetBitmap(down);
m_disasm.InitStorage(LINES,256);
m_tabs.InsertItem(0,"main cpu");	
m_tabs.InsertItem(1,"sound cpu");	
currentcpu = 0;
changecpu();
refresh();
	// TODO: Add extra initialization here
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CDebugger::OnSelchangeCputabs(NMHDR* /*pNMHDR*/, LRESULT* pResult) 
{
	// TODO: Add your control notification handler code here
currentcpu = (u8)m_tabs.GetCurSel();
changecpu();
refresh();
	*pResult = 0;
}

void CDebugger::OnRefresh() 
{
	// TODO: Add your control notification handler code here
refresh();
}

void CDebugger::OnUp() 
{
	// TODO: Add your control notification handler code here
pc--;
refresh();
}

void CDebugger::OnDown() 
{
	// TODO: Add your control notification handler code here
pc++;
refresh();
}

extern u8 vblank;

void CDebugger::OnStep() 
{
	// TODO: Add your control notification handler code here
char *p,*stop;
int base = 10;
u32 num,i;

UpdateData(TRUE);
p = m_stepcycles.GetBuffer(512);
if(strcmp(p,"") != 0)
	{
	if(*p == '$')
		{
		p++;
		base = 16;
		}
	else if((*p == '0' && (*(p+1) == 'x') || (*p == '0' && *(p+1) == 'X')))
		{
		p += 2;
		base = 16;
		}
	num = strtol(p,&stop,base);
	for(i=0;i<num;)
		{
		u32 poop;

		poop = exec6502(1);
		if(poop & BAD_OPCODE)
			{
			emumessage("%s: bad opcode at %04x",currentcpu==0?"main":"sound",cpu->pc);
			break;
			}
		else
			{
			i += poop;
			if(currentcpu == 0)
				{
				cpu1cycles += poop;
				if(cpu1cycles > cpu1screen)
					vblank = 0;
				else
					vblank = 0x80;
				if(cpu1cycles > cpu1frame)
					cpu1cycles -= cpu1frame;
				}
			else
				{
				cpu2cycles += poop;
				if(cpu2cycles > cpu2frame)
					cpu2cycles -= cpu2frame;
				}
			}
		}
	pc = cpu->pc;
	refresh();
	}
}

extern u8 *cpumem[];

void CDebugger::OnReset() 
{
	// TODO: Add your control notification handler code here
reset6502();
pc = cpu->pc;
refresh();
cpu1cycles = 0;
cpu2cycles = 0;
}

BOOL CDebugger::OnCommand(WPARAM wParam, LPARAM lParam) 
{
	// TODO: Add your specialized code here and/or call the base class
if(LOWORD(wParam) == IDCANCEL)
	{
	theApp.OnViewDebugger();
	return(1);
	}
else
	return CDialog::OnCommand(wParam, lParam);
}

void CDebugger::OnClose() 
{
	// TODO: Add your message handler code here and/or call default
	
	CDialog::OnClose();
theApp.OnViewDebugger();
}

void CDebugger::OnOK() 
{
	// TODO: Add extra validation here
	
	CDialog::OnOK();
theApp.OnViewDebugger();
}

void CDebugger::OnBreak() 
{
	// TODO: Add your control notification handler code here
char *p,*stop;
int base = 10;
u16 lastpc,breakpc;

UpdateData(TRUE);
p = m_breakaddr.GetBuffer(512);
if(strcmp(p,"") != 0)
	{
	if(*p == '$')
		{
		p++;
		base = 16;
		}
	else if((*p == '0' && (*(p+1) == 'x') || (*p == '0' && *(p+1) == 'X')))
		{
		p += 2;
		base = 16;
		}
	breakpc = (u16)strtol(p,&stop,base);
	lastpc = 0;
	for(;;)
		{
/*		if(exec6502(1) & BAD_OPCODE)
			emumessage("%s: bad opcode at %04x",currentcpu==0?"main":"sound",cpu->pc);
		else if(cpu->pc == breakpc)
			break;
		else if(lastpc < breakpc && cpu->pc > breakpc) //skipped over the break address
			break;*/
		u32 poop;

		poop = exec6502(1);
		if(poop & BAD_OPCODE)
			{
			emumessage("%s: bad opcode at %04x",currentcpu==0?"main":"sound",cpu->pc);
			break;
			}
		else
			{
			if(currentcpu == 0)
				{
				cpu1cycles += poop;
				if(cpu1cycles > cpu1screen)
					vblank = 0;
				else
					vblank = 0x80;
				if(cpu1cycles > cpu1frame)
					cpu1cycles -= cpu1frame;
				}
			else
				{
				cpu2cycles += poop;
				if(cpu2cycles > cpu2frame)
					cpu2cycles -= cpu2frame;
				}
			if(cpu->pc == breakpc)
				break;
			else if(lastpc < breakpc && /*cpu->pc*/(lastpc+5) > breakpc) //skipped over the break address
				break;
			lastpc = cpu->pc;
			if(GetAsyncKeyState(VK_F1))
				break;
			}
		}
	pc = cpu->pc;
	emumessage("break at %04X (wanted %04X)\n",pc,breakpc);
	refresh();
	}
}

void CDebugger::OnJump() 
{
	// TODO: Add your control notification handler code here
char *p,*stop;
int base = 10;

UpdateData(TRUE);
p = m_jumpaddr.GetBuffer(512);
if(strcmp(p,"") != 0)
	{
	if(*p == '$')
		{
		p++;
		base = 16;
		}
	else if((*p == '0' && (*(p+1) == 'x') || (*p == '0' && *(p+1) == 'X')))
		{
		p += 2;
		base = 16;
		}
	pc = (u16)strtol(p,&stop,base);
	refresh();
	}
}

void CDebugger::OnIrq() 
{
	// TODO: Add your control notification handler code here
irq6502();
}

void CDebugger::OnNmi() 
{
	// TODO: Add your control notification handler code here
nmi6502();
}
