#include <windows.h>
#include <ddraw.h>
#include "stdafx.h"
#include "blitter.h"
/*********************************************************************************************/
gdiblitter::gdiblitter()
{
palette = 0;
screen = 0;
}
/*********************************************************************************************/
gdiblitter::~gdiblitter()
{
kill();
}
/*********************************************************************************************/
int gdiblitter::init(int width,int height,int colors,HWND wnd)
{
HWND hdesktopwnd = GetDesktopWindow();
HDC hdesktopdc = GetDC(hdesktopwnd);

ASSERT(this);
hwnd = wnd;
w = width;
h = height;
numcolors = colors;
bpp = GetDeviceCaps(hdesktopdc,BITSPIXEL);
DeleteDC(hdesktopdc);
memset(&bmpinfo,0,sizeof(bmpinfo));
bmpinfo.bmiHeader.biSize        = sizeof(BITMAPINFOHEADER);
bmpinfo.bmiHeader.biBitCount    = (u16)bpp;
bmpinfo.bmiHeader.biWidth       = w;
bmpinfo.bmiHeader.biHeight      = -h;
bmpinfo.bmiHeader.biPlanes      = 1;	
bmpinfo.bmiHeader.biCompression = BI_RGB;
bmpinfo.bmiHeader.biSizeImage   = 0;
switch(bpp)
	{
	case 16:
		if((hbitmap = CreateDIBSection(0,&bmpinfo,DIB_RGB_COLORS,&dest,0,0)) == 0)
			{
			MessageBox(0,"CreateDIBSection","gdiblitter::init error",MB_OK|MB_ICONERROR);
			return(0);
			}
		palette = new u16[colors];
		break;
	case 32:
		if((hbitmap = CreateDIBSection(0,&bmpinfo,DIB_RGB_COLORS,&dest,0,0)) == 0)
			{
			MessageBox(0,"CreateDIBSection","gdiblitter::init error",MB_OK|MB_ICONERROR);
			return(0);
			}
		palette = new u32[colors];
		break;
	default:
		MessageBox(0,"try using 16 or 32 bit desktop bpp","gdiblitter::init error",MB_OK|MB_ICONERROR);
		return(0);
	}
memset(palette,0,colors * ((bpp & ~1) / 8));
screen = new u8[w*h];
hdc = GetDC(hwnd);
resize();
return(bpp);
}
/*********************************************************************************************/
void gdiblitter::kill()
{
ASSERT(this);
DeleteObject(dest);
DeleteObject(hbitmap);
ReleaseDC(hwnd,hdc);
if(palette)
	{
	delete[] palette;
	palette = 0;
	}
if(screen)
	{
	delete[] screen;
	screen = 0;
	}
}
/*********************************************************************************************/
void gdiblitter::setcolor(int num,u8 r,u8 g,u8 b)
{
ASSERT(this);
if(bpp == 16)
	((u16*)palette)[num] = (u16)(((r >> 3) << 10) | ((g >> 3) << 5) | (b >> 3));
else if(bpp == 17)
	((u16*)palette)[num] = (u16)(((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3));
else if(bpp == 32)
	((u32*)palette)[num] = (u32)((r << 16) | (g << 8) | b);
}
/*********************************************************************************************/
void gdiblitter::getcolor(int num,u8 *r,u8 *g,u8 *b)
{
ASSERT(this);
if(bpp == 16)
	{
	*r = (u8)(((((u16*)palette)[num] >> 10) & 0x1f) << 3);
	*g = (u8)(((((u16*)palette)[num] >> 5) & 0x1f) << 3);
	*b = (u8)(((((u16*)palette)[num] >> 0) & 0x1f) << 3);
	}
else if(bpp == 17)
	{
	*r = (u8)(((((u16*)palette)[num] >> 11) & 0x1f) << 3);
	*g = (u8)(((((u16*)palette)[num] >> 5) & 0x3f) << 2);
	*b = (u8)(((((u16*)palette)[num] >> 0) & 0x1f) << 3);
	}
else if(bpp == 32)
	{
	*r = (u8)(((u32*)palette)[num] >> 16);
	*g = (u8)(((u32*)palette)[num] >> 8);
	*b = (u8)(((u32*)palette)[num] >> 0);
	}
else
	{
	*r = 0;
	*g = 0;
	*b = 0;
	}
}
/*********************************************************************************************/
void gdiblitter::resize()
{
ASSERT(this);
GetClientRect(hwnd,&rect);
}
/*********************************************************************************************/
u8 *gdiblitter::framestart()
{
ASSERT(this);
return(screen);
}
/*********************************************************************************************/
void gdiblitter::framestop()
{
int x,y;

ASSERT(this);
if(bpp & 16)
	{
	u8 *s = screen;
	u16 *d = (u16*)dest;

	s += w;
	for(y=0;y<h;y++)
		{
		u8 *ss = s;

		for(x=0;x<w;x++)
			{
			*d++ = ((u16*)palette)[*ss];
			ss += w;
			}
		s--;
		}
	}
else if(bpp & 32)
	{
	u8 *s = screen;
	u32 *d = (u32*)dest;

	s += w;
	for(y=0;y<h;y++)
		{
		u8 *ss = s;

		for(x=0;x<w;x++)
			{
			*d++ = ((u32*)palette)[*ss];
			ss += w;
			}
		s--;
		}
	}
StretchDIBits(hdc,0,0,rect.right,rect.bottom,0,0,w,h,dest,&bmpinfo,DIB_RGB_COLORS,SRCCOPY);
}
/*********************************************************************************************/
void gdiblitter::getpalette(void **dest)
{
ASSERT(this);
switch(bpp)
	{
	case 16:*dest = new u16[numcolors];break;
	case 32:*dest = new u32[numcolors];break;
	default:*dest = 0;break;
	}
memcpy(*dest,palette,numcolors * (bpp / 8));
}
/*********************************************************************************************/
void gdiblitter::setpalette(void *src)
{
ASSERT(this);
memcpy(palette,src,numcolors * (bpp / 8));
}
/*********************************************************************************************/
dxblitter::dxblitter()
{
lpdd = 0;
palette = 0;
screen = 0;
}
/*********************************************************************************************/
dxblitter::~dxblitter()
{
kill();
}
/*********************************************************************************************/
int dxblitter::init(int width,int height,int colors,HWND wnd)
{
ASSERT(this);
hwnd = wnd;
w = width;
h = height;
numcolors = colors;
if(DirectDrawCreateEx(0,(VOID**)&lpdd,IID_IDirectDraw7,0))
	return(0);
if(lpdd->SetCooperativeLevel(hwnd,DDSCL_NORMAL) != DD_OK)
	{
	MessageBox(0,"SetCooperativeLevel","dxblitter::init error",MB_OK|MB_ICONERROR);
	kill();
	return(0);
	}
memset(&ddsd,0,sizeof(DDSURFACEDESC2));
ddsd.dwSize = sizeof(DDSURFACEDESC2);
ddsd.dwFlags = DDSD_CAPS;
ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
if(lpdd->CreateSurface(&ddsd,&lpddsprimary,0) != DD_OK)
	{
	MessageBox(0,"CreateSurface","dxblitter::init error",MB_OK|MB_ICONERROR);
	kill();
	return(0);
	}
if(lpdd->CreateClipper(0,&lpclipper, NULL) != DD_OK)
	{
	MessageBox(0,"CreateClipper","dxblitter::init error",MB_OK|MB_ICONERROR);
	kill();
	return(0);
	}
lpclipper->SetHWnd(0,hwnd);
lpddsprimary->SetClipper(lpclipper);
lpclipper->Release();
memset(&ddsd,0,sizeof(DDSURFACEDESC2));
ddsd.dwSize = sizeof(DDSURFACEDESC2);
ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
ddsd.dwWidth = w;
ddsd.dwHeight = h;
ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
if(lpdd->CreateSurface(&ddsd,&lpddsback,0) != DD_OK)
	{
	MessageBox(0,"CreateSurface","dxblitter::init error",MB_OK|MB_ICONERROR);
	kill();
	return(0);
	}
memset(&ddsd,0,sizeof(DDSURFACEDESC2));
ddsd.dwSize = sizeof(ddsd);
if(lpddsback->GetSurfaceDesc(&ddsd) != DD_OK)
	{
	MessageBox(0,"GetSurfaceDesc","dxblitter::init error",MB_OK|MB_ICONERROR);
	kill();
	return(0);
	}
bpp = (byte)ddsd.ddpfPixelFormat.dwRGBBitCount;
if(ddsd.ddpfPixelFormat.dwRBitMask == 0xf800 && bpp == 16)
	bpp++;
switch(bpp & ~1)
	{
	case 16:palette = new u16[colors];break;
	case 32:palette = new u32[colors];break;
	default:
		MessageBox(0,"try using 16 or 32 bit desktop bpp","dxblitter::init error",MB_OK|MB_ICONERROR);
		kill();
		return(0);
	}
screen = new u8[w*h];
memset(palette,0,colors * ((bpp & ~1) / 8));
resize();
return(bpp & ~1);
}
/*********************************************************************************************/
void dxblitter::kill()
{
ASSERT(this);
if(lpdd != 0)
	{
	lpdd->SetCooperativeLevel(hwnd,DDSCL_NORMAL);
	if(lpddsback != 0)
		{
		lpddsback->Release();
		lpddsback = 0;
		}
	if(lpddsprimary != 0)
		{
		lpddsprimary->Release();
		lpddsprimary = 0;
		}
	lpdd->Release();
	lpdd = 0;
	}
if(palette)
	{
	delete[] palette;
	palette = 0;
	}
if(screen)
	{
	delete[] screen;
	screen = 0;
	}
}
/*********************************************************************************************/
void dxblitter::setcolor(int num,u8 r,u8 g,u8 b)
{
ASSERT(this);
if(bpp == 16)
	((u16*)palette)[num] = (u16)(((r >> 3) << 10) | ((g >> 3) << 5) | (b >> 3));
else if(bpp == 17)
	((u16*)palette)[num] = (u16)(((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3));
else if(bpp == 32)
	((u32*)palette)[num] = (u32)((r << 16) | (g << 8) | b);
}
/*********************************************************************************************/
void dxblitter::getcolor(int num,u8 *r,u8 *g,u8 *b)
{
ASSERT(this);
if(bpp == 16)
	{
	*r = (u8)(((((u16*)palette)[num] >> 10) & 0x1f) << 3);
	*g = (u8)(((((u16*)palette)[num] >> 5) & 0x1f) << 3);
	*b = (u8)(((((u16*)palette)[num] >> 0) & 0x1f) << 3);
	}
else if(bpp == 17)
	{
	*r = (u8)(((((u16*)palette)[num] >> 11) & 0x1f) << 3);
	*g = (u8)(((((u16*)palette)[num] >> 5) & 0x3f) << 2);
	*b = (u8)(((((u16*)palette)[num] >> 0) & 0x1f) << 3);
	}
else if(bpp == 32)
	{
	*r = (u8)(((u32*)palette)[num] >> 16);
	*g = (u8)(((u32*)palette)[num] >> 8);
	*b = (u8)(((u32*)palette)[num] >> 0);
	}
else
	{
	*r = 0;
	*g = 0;
	*b = 0;
	}
}
/*********************************************************************************************/
void dxblitter::resize()
{
RECT rectClient,rectWindow;
int w,h;

ASSERT(this);
GetWindowRect(hwnd,&rectWindow);
GetClientRect(hwnd,&rectClient);
w = (rectWindow.right - rectWindow.left) - rectClient.right;
h = (rectWindow.bottom - rectWindow.top) - rectClient.bottom;
w /= 2;
GetWindowRect(hwnd,&rect);
rect.top += h - w;
rect.left += w;
rect.right -= w;
w++;
rect.bottom -= w;
}
/*********************************************************************************************/
u8 *dxblitter::framestart()
{
int ret;

ASSERT(this);
if((ret = lpddsback->Lock(0,&ddsd,DDLOCK_WAIT,0)) == DD_OK)
	{
	dest = ddsd.lpSurface;
	return(screen);
	}
else
	{
	switch(ret)
		{
		case DDERR_INVALIDOBJECT:MessageBox(0,"DDERR_INVALIDOBJECT","dxvideo_blit: error",MB_OK|MB_ICONERROR);break;
		case DDERR_INVALIDPARAMS:MessageBox(0,"DDERR_INVALIDPARAMS","dxvideo_blit: error",MB_OK|MB_ICONERROR);break;
		case DDERR_OUTOFMEMORY:MessageBox(0,"DDERR_OUTOFMEMORY","dxvideo_blit: error",MB_OK|MB_ICONERROR);break;
		case DDERR_SURFACEBUSY:MessageBox(0,"DDERR_SURFACEBUSY","dxvideo_blit: error",MB_OK|MB_ICONERROR);break;
		case DDERR_SURFACELOST:MessageBox(0,"DDERR_SURFACELOST","dxvideo_blit: error",MB_OK|MB_ICONERROR);break;
		case DDERR_WASSTILLDRAWING:MessageBox(0,"DDERR_WASSTILLDRAWING","dxvideo_blit: error",MB_OK|MB_ICONERROR);break;
		}
	}
return(0);
}
/*********************************************************************************************/
void dxblitter::framestop()
{
int x,y;

ASSERT(this);
if(bpp & 16)
	{
	u8 *s = screen;
	u16 *d = (u16*)dest;

	s += w;
	for(y=0;y<h;y++)
		{
		u8 *ss = s;

		for(x=0;x<w;x++)
			{
			*d++ = ((u16*)palette)[*ss];
			ss += w;
			}
		s--;
		}
	}
else if(bpp & 32)
	{
	u8 *s = screen;
	u32 *d = (u32*)dest;

	s += w;
	for(y=0;y<h;y++)
		{
		u8 *ss = s;

		for(x=0;x<w;x++)
			{
			*d++ = ((u32*)palette)[*ss];
			ss += w;
			}
		s--;
		}
	}
lpddsback->Unlock(0);
if(lpddsprimary->Blt(&rect,lpddsback,0,DDBLT_WAIT,0) == DDERR_SURFACELOST)
	lpddsprimary->Restore();
}
/*********************************************************************************************/
void dxblitter::getpalette(void **dest)
{
ASSERT(this);
switch(bpp & ~1)
	{
	case 16:*dest = new u16[numcolors];break;
	case 32:*dest = new u32[numcolors];break;
	default:*dest = 0;break;
	}
memcpy(*dest,palette,numcolors * (bpp / 8));
}
/*********************************************************************************************/
void dxblitter::setpalette(void *src)
{
ASSERT(this);
memcpy(palette,src,numcolors * (bpp / 8));
}
/*********************************************************************************************/
